#include <stdlib.h>
#include <time.h>
#include "FloatPixelUtils.h"
#include "OrderFilter.h"


Image *OrderFilter::naiveMedian(Image &img,
                                   int halfWidth, int halfHeight)
{
	int x,y,curX,curY,startX,endX,startY,endY,index;
	int minX, maxX, minY, maxY;
  float *data = new float [(halfWidth*2+1)*(halfHeight*2+1)];
  Image *res = new Image(img);

	minY= halfHeight;
	maxY = img.getHeight() - halfHeight;
	minX = halfWidth;
	maxX = img.getWidth() - halfWidth;

	for(y=minY; y<maxY; y++) {
		startY = y - halfHeight;
		endY = y + halfHeight;
		for(x=minX; x<maxX; x++) {
			startX = x - halfWidth;
			endX = x + halfWidth;
			index=0;
			for(curY = startY; curY<=endY; curY++) {
        for(curX = startX; curX<=endX; curX++) {
          data[index++] = img.Pix(curX,curY);
        }
			}
      qsort(data,index,sizeof(float),FloatPixelUtils::compareGrayValues);
      res->Pix(x,y) = data[index/2];
    }
  }
  delete [] data;
  return(res);
}
/*****************************************************************************/
Image *OrderFilter::hoareSelectK(Image &img, float percentage,
                                   int halfWidth, int halfHeight)
{
	int x,y,curX,curY,startX,endX,startY,endY,index;
	int minX, maxX, minY, maxY;
  float *data = new float [(halfWidth*2+1)*(halfHeight*2+1)];
  Image *res = new Image(img);

	minY= halfHeight;
	maxY = img.getHeight() - halfHeight;
	minX = halfWidth;
	maxX = img.getWidth() - halfWidth;


	for(y=minY; y<maxY; y++) {
		startY = y - halfHeight;
		endY = y + halfHeight;
		for(x=minX; x<maxX; x++) {
			startX = x - halfWidth;
			endX = x + halfWidth;
			index=0;
			for(curY = startY; curY<=endY; curY++) {
				for(curX = startX; curX<=endX; curX++) {
					data[index++] = img.Pix(curX,curY);
				}
			}
			res->Pix(x,y) = hoareSelect(data,0,index-1,(int)(percentage*(index-1) + 0.5));
		}
	}
	delete [] data;
	return(res);
}
/*****************************************************************************/
float OrderFilter::hoareSelect(float *data, int start, int end, int k)
{
  int pivotIndex, i, j;
  float pivotValue, tmp;
       //seed the random number generator
  srand((unsigned)time(NULL));

  while(start < end) {
          //select a random pivot 
    pivotIndex = (int)(rand() * ((float)(end-start)/(float)RAND_MAX) + start);
    pivotValue = data[pivotIndex];
             //sort the array into two sub arrays around the pivot value
    i = start;
    j = end;
    while(i <= j) {
      while(data[i] < pivotValue)
        i++;
      while(data[j] > pivotValue)
        j--;
      if(i<=j) {
        tmp = data[i];
        data[i] = data[j];
        data[j] = tmp;
        i++;
        j--;
      }
    } 
          //continue search in left sub array
    if(j < k)
      start = i;
    if(k < i) //continue search in right sub array
      end = j;
  }
  return(data[k]);
}
/*****************************************************************************/
Image *OrderFilter::histoSelectK(Image &img, float percentage,
                                   int halfWidth, int halfHeight,
																	 int minGrayLevel, int maxGrayLevel)
{
	int x,y,curX,curY,startX,endX,startY,endY;
	int i, k, curNumber, minX, maxX, minY, maxY;
	int *histo = new int[maxGrayLevel - minGrayLevel + 1];
	histo -= minGrayLevel*sizeof(int);
	
	k = (int)(percentage*(2*halfHeight+1)*(2*halfWidth+1) + 0.5);

  Image *res = new Image(img);

	minY= halfHeight;
	maxY = img.getHeight() - halfHeight;
	minX = halfWidth;
	maxX = img.getWidth() - halfWidth;
	for(y=minY; y<maxY; y++) {
		startY = y - halfHeight;
		endY = y + halfHeight;
		for(x=minX; x<maxX; x++) {
			memset(&(histo[minGrayLevel]),'\0',(maxGrayLevel - minGrayLevel + 1)*sizeof(int));
			startX = x - halfWidth;
			endX = x + halfWidth;
			for(curY = startY; curY<=endY; curY++) {
				for(curX = startX; curX<=endX; curX++) {
					histo[(int)img.Pix(curX,curY)]++;
				}
			}           //we created the histogram in the current window, now get the k'th
			            //smallest element (according to the given percentage)
			i= minGrayLevel;
			curNumber = histo[i];
			while(curNumber < k) {
				i++;
				curNumber+= histo[i];
			}
			res->Pix(x,y) = i;
		}
	}
	histo+= minGrayLevel*sizeof(int);
	delete [] histo;
	return(res);
}
/*****************************************************************************/
 Image *OrderFilter::median3X3(Image &img)
{
	int x,y,curY;
  int width = img.getWidth();
  int height = img.getHeight();
  float data[9];
  Image *res = new Image(img);

	                 //main image part [1,1]..[width-2,height-2]
	for(y=1; y<height-1; y++) {
		for(x=1; x<width-1; x++) {
			curY = y-1;
			data[0] = img.Pix(x-1,curY); 
			data[1] = img.Pix(x,curY);
			data[2] = img.Pix(x+1,curY);
			
			curY = y;
			data[3] = img.Pix(x-1,curY); 
			data[4] = img.Pix(x,curY);
			data[5] = img.Pix(x+1,curY);
			
			curY = y+1;
			data[6] = img.Pix(x-1,curY); 
			data[7] = img.Pix(x,curY);
			data[8] = img.Pix(x+1,curY);
			res->Pix(x,y) = median9(data);
		}
	}
	return(res);
 }
/*****************************************************************************/
 Image *OrderFilter::median5X5(Image &img)
{
	int x,y,curY;
  int width = img.getWidth();
  int height = img.getHeight();
  float data[25];
  Image *res = new Image(img);


	for(y=2; y<height-2; y++) {
		for(x=2; x<width-2; x++) {

			curY= y-2;
			data[0] = img.Pix(x-2,curY);  
			data[1] = img.Pix(x-1,curY);
			data[2] = img.Pix(x,curY);
			data[3] = img.Pix(x+1,curY); 
			data[4] = img.Pix(x+2,curY);

			curY= y-1;
			data[5] = img.Pix(x-2,curY);  
			data[6] = img.Pix(x-1,curY);
			data[7] = img.Pix(x,curY);
			data[8] = img.Pix(x+1,curY); 
			data[9] = img.Pix(x+2,curY);

			curY= y;
			data[10] = img.Pix(x-2,curY);  
			data[11] = img.Pix(x-1,curY);
			data[12] = img.Pix(x,curY);
			data[13] = img.Pix(x+1,curY); 
			data[14] = img.Pix(x+2,curY);

			curY= y+1;
			data[15] = img.Pix(x-2,curY);  
			data[16] = img.Pix(x-1,curY);
			data[17] = img.Pix(x,curY);
			data[18] = img.Pix(x+1,curY); 
			data[19] = img.Pix(x+2,curY);

			curY= y+2;
			data[20] = img.Pix(x-2,curY);  
			data[21] = img.Pix(x-1,curY);
			data[22] = img.Pix(x,curY);
			data[23] = img.Pix(x+1,curY); 
			data[24] = img.Pix(x+2,curY);

			res->Pix(x,y) = median25(data);
		}
	}
	return(res);
 }
/*****************************************************************************/
float OrderFilter::median3(float *data)
{
    pixSort(data[0],data[1]); pixSort(data[1],data[2]); pixSort(data[0],data[1]);
    return(data[1]);
}
/*****************************************************************************/
float OrderFilter::median5(float *data)
{
	pixSort(data[0],data[1]); pixSort(data[3],data[4]); pixSort(data[0],data[3]);
	pixSort(data[1],data[4]); pixSort(data[1],data[2]); pixSort(data[2],data[3]);
	pixSort(data[1],data[2]); 
	return(data[2]);
}
/*****************************************************************************/
float OrderFilter::median7(float *data)
{    
	pixSort(data[0], data[5]); pixSort(data[0], data[3]); pixSort(data[1], data[6]);
	pixSort(data[2], data[4]); pixSort(data[0], data[1]); pixSort(data[3], data[5]);
	pixSort(data[2], data[6]); pixSort(data[2], data[3]); pixSort(data[3], data[6]);
	pixSort(data[4], data[5]); pixSort(data[1], data[4]); pixSort(data[1], data[3]);
	pixSort(data[3], data[4]); 
	return (data[3]);
}
/*****************************************************************************/
float OrderFilter::median9(float *data)
{
	pixSort(data[1], data[2]); pixSort(data[4], data[5]); pixSort(data[7], data[8]);
	pixSort(data[0], data[1]); pixSort(data[3], data[4]); pixSort(data[6], data[7]);
	pixSort(data[1], data[2]); pixSort(data[4], data[5]); pixSort(data[7], data[8]);
	pixSort(data[0], data[3]); pixSort(data[5], data[8]); pixSort(data[4], data[7]);
	pixSort(data[3], data[6]); pixSort(data[1], data[4]); pixSort(data[2], data[5]);
	pixSort(data[4], data[7]); pixSort(data[4], data[2]); pixSort(data[6], data[4]);
	pixSort(data[4], data[2]); 
	return(data[4]) ;
}
/*****************************************************************************/
 float OrderFilter::median25(float *data)
{
	pixSort(data[0], data[1]);   pixSort(data[3], data[4]);   pixSort(data[2], data[4]);
	pixSort(data[2], data[3]);   pixSort(data[6], data[7]);   pixSort(data[5], data[7]);
	pixSort(data[5], data[6]);   pixSort(data[9], data[10]);  pixSort(data[8], data[10]);
	pixSort(data[8], data[9]);   pixSort(data[12], data[13]); pixSort(data[11], data[13]);
	pixSort(data[11], data[12]); pixSort(data[15], data[16]); pixSort(data[14], data[16]);
	pixSort(data[14], data[15]); pixSort(data[18], data[19]); pixSort(data[17], data[19]);
	pixSort(data[17], data[18]); pixSort(data[21], data[22]); pixSort(data[20], data[22]);
	pixSort(data[20], data[21]); pixSort(data[23], data[24]); pixSort(data[2], data[5]);
	pixSort(data[3], data[6]);   pixSort(data[0], data[6]);   pixSort(data[0], data[3]);
	pixSort(data[4], data[7]);   pixSort(data[1], data[7]);   pixSort(data[1], data[4]);
	pixSort(data[11], data[14]); pixSort(data[8], data[14]);  pixSort(data[8], data[11]);
	pixSort(data[12], data[15]); pixSort(data[9], data[15]);  pixSort(data[9], data[12]);
	pixSort(data[13], data[16]); pixSort(data[10], data[16]); pixSort(data[10], data[13]);
	pixSort(data[20], data[23]); pixSort(data[17], data[23]); pixSort(data[17], data[20]);
	pixSort(data[21], data[24]); pixSort(data[18], data[24]); pixSort(data[18], data[21]);
	pixSort(data[19], data[22]); pixSort(data[8], data[17]);  pixSort(data[9], data[18]);
	pixSort(data[0], data[18]);  pixSort(data[0], data[9]);   pixSort(data[10], data[19]);
	pixSort(data[1], data[19]);  pixSort(data[1], data[10]);  pixSort(data[11], data[20]);
	pixSort(data[2], data[20]);  pixSort(data[2], data[11]);  pixSort(data[12], data[21]);
	pixSort(data[3], data[21]);  pixSort(data[3], data[12]);  pixSort(data[13], data[22]);
	pixSort(data[4], data[22]);  pixSort(data[4], data[13]);  pixSort(data[14], data[23]);
	pixSort(data[5], data[23]);  pixSort(data[5], data[14]);  pixSort(data[15], data[24]);
	pixSort(data[6], data[24]);  pixSort(data[6], data[15]);  pixSort(data[7], data[16]);
	pixSort(data[7], data[19]);  pixSort(data[13], data[21]); pixSort(data[15], data[23]);
	pixSort(data[7], data[13]);  pixSort(data[7], data[15]);  pixSort(data[1], data[9]);
	pixSort(data[3], data[11]);  pixSort(data[5], data[17]);  pixSort(data[11], data[17]);
	pixSort(data[9], data[17]);  pixSort(data[4], data[10]);  pixSort(data[6], data[12]);
	pixSort(data[7], data[14]);  pixSort(data[4], data[6]);   pixSort(data[4], data[7]);
	pixSort(data[12], data[14]); pixSort(data[10], data[14]); pixSort(data[6], data[7]);
	pixSort(data[10], data[12]); pixSort(data[6], data[10]);  pixSort(data[6], data[17]);
	pixSort(data[12], data[17]); pixSort(data[7], data[17]);  pixSort(data[7], data[10]);
	pixSort(data[12], data[18]); pixSort(data[7], data[12]);  pixSort(data[10], data[18]);
	pixSort(data[12], data[20]); pixSort(data[10], data[20]); pixSort(data[10], data[12]);
		
  return (data[12]);
}
