#ifndef _RUN_MED_HISTO_H_
#define _RUN_MED_HISTO_H_

#include "Image.h"

/*
 * Implementation of running median using a histogram.
 *
 * NOTE: This median works only with integral values (although the image data is
 * ----  of type float).
 *
 * The algorithm simply traverses the image from left to right top to bottom and
 * updates a histogram which enables finding the median in constant time for every
 * pixel.
 *
 * Author: Ziv Yaniv (zivy@cs.huji.ac.il)
 */

class RunMedHisto {

public:
	    /*
       * Median using a window of size 2*halfWidth+1 X 2*halfHeight+1
	     * Gray values in the image are in [minGrayLevel, maxGrayLevel].
	     * I assume that minGrayLevel<maxGrayLevel. The values may be negative.
			 */
	RunMedHisto(int halfWidth, int halfHeight, 
		          int minGrayLevel, int maxGrayLevel); 
	~RunMedHisto();
   /*
    * Apply the median filter to the given image.
    * Freeing the memory used by the returned image is the users 
    * responsibility.
    */
	Image *apply(Image &img);

private:
	void initData(Image &img);
	void initNextRow(Image &img, int x, int y);
	inline int getMedian();

	int *_histo;
	int _halfWidth;
	int _halfHeight;
	int _minGrayLevel;
	int _maxGrayLevel;
	int _medianNumber;
};

#endif //_RUN_MED_HISTO_H_


